#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "tinyxml.h"

#include "outstreams.hpp"
#include "config.hpp"
#include "coordinates.hpp"
#include "SC2Map.hpp"



float SC2Map::defaultMineralAmount = 1500.0f;
float SC2Map::defaultGeyserAmount  = 2500.0f;



void SC2Map::processPlacedObject( ObjectMode objMode, TiXmlElement* object )
{
  char* objPoint;
  char* objUnit;
  char* objDoodad;
  char* attType;
  char* attPosition;
  char* attUnitType;
  char* attResources;

  switch( objMode )
  {
    case OBJMODE_BETAPH1_v26:
      objPoint     = "Point";
      objUnit      = "Unit";
      objDoodad    = "Doodad";
      attType      = "type";
      attPosition  = "position";
      attUnitType  = "unitType";
      attResources = "resources";
    break;


    case OBJMODE_BETAPH2_v26:
      objPoint     = "ObjectPoint";
      objUnit      = "ObjectUnit";
      objDoodad    = "ObjectDoodad";
      attType      = "Type";
      attPosition  = "Position";
      attUnitType  = "UnitType";
      attResources = "Resources";
    break;


    default:
      printError( "Unknown object processing mode?\n" );
      exit( -1 );
  }


  // Points might be a start location
  if( strcmp( object->Value(), objPoint ) == 0 )
  {
    const char* strType = object->Attribute( attType );
    if( strcmp( strType, "StartLoc" ) != 0 )
    {
      return;
    }

    const char* strPosition = object->Attribute( attPosition );
    float mx, my, mz;
    sscanf( strPosition, "%f,%f,%f", &mx, &my, &mz );

    StartLoc* sl = new StartLoc;
    sl->loc.mSet( mx, my );
    startLocs.push_back( sl );
    return;
  }


  // doodads might have a footprint, don't worry about scaled
  // doodads--their footprint is the same
  if( strcmp( object->Value(), objDoodad ) == 0 )
  {
    const char* strType     = object->Attribute( attType );

    const char* strPosition = object->Attribute( attPosition );
    float mx, my, mz;
    sscanf( strPosition, "%f,%f,%f", &mx, &my, &mz );
    point c;
    c.mSet( mx, my );

    string type( "doodad" );
    string name( strType );

    applyFootprint( &c, &type, &name );
  }



  // units come in three flavors:
  //   resource - can be mined, has a footprint
  //   destruct - can be destroyed, has a footprint
  //   unit     - may have a footprint

  if( strcmp( object->Value(), objUnit ) == 0 )
  {
    // every unit has a position (TRUE?)
    const char* strPosition = object->Attribute( attPosition );
    float mx, my, mz;
    sscanf( strPosition, "%f,%f,%f", &mx, &my, &mz );
    point c;
    c.mSet( mx, my );

    // every unit has a unitType (TRUE?)
    const char* strType = object->Attribute( attUnitType );

    // might be used by resources below
    const char* strResources = object->Attribute( attResources );



    // an object may have more than one type of footprint,
    // all of which will be applied
    string name( strType );

    string type( "destruct" );
    applyFootprint( &c, &type, &name );

    type.assign( "unit" );
    applyFootprint( &c, &type, &name );

    type.assign( "resource" );
    applyFootprint( &c, &type, &name );

    type.assign( "nobuild" );
    applyFootprint( &c, &type, &name );



    if( strcmp( strType, "Observatory" ) == 0 )
    {
      Watchtower* wt = new Watchtower;
      wt->loc.mSet( mx, my );
      wt->range = 22.0f;
      watchtowers.push_back( wt );
      return;
    }


    if( strcmp( strType, "MineralField" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = MINERALS;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = defaultMineralAmount;
      }
      resources.push_back( r );
      return;
    }

    if( strcmp( strType, "RichMineralField"      ) == 0 ||
        strcmp( strType, "HighYieldMineralField" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = MINERALS_HY;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = defaultMineralAmount;
      }
      resources.push_back( r );
      return;
    }

    if( strcmp( strType, "VespeneGeyser"       ) == 0 ||
        strcmp( strType, "SpacePlatformGeyser" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = VESPENEGAS;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = defaultGeyserAmount;
      }
      resources.push_back( r );
      return;
    }

    if( strcmp( strType, "RichVespeneGeyser" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = VESPENEGAS_HY;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = defaultGeyserAmount;
      }
      resources.push_back( r );
      return;
    }
  }
}



void SC2Map::applyFootprint( point* c, string* type, string* name )
{
  Footprint* foot = NULL;;

  map<string, Footprint*>* innerMap;


  // first check local user config
  innerMap = configUserLocal.type2name2foot[*type];
  if( innerMap != NULL )
  {
    foot = (*innerMap)[*name];
  }


  // if nothing there, check global
  if( foot == NULL )
  {
    innerMap = configUserGlobal.type2name2foot[*type];
    if( innerMap != NULL )
    {
      foot = (*innerMap)[*name];
    }
  }


  // note: there are NO FOOTPRINTS in the global internal
  // config, so if we don't have one by now, just quit
  if( foot == NULL )
  {
    return;
  }

  for( list<int>::iterator itr = foot->relativeCoordinates.begin();
       itr != foot->relativeCoordinates.end();
       ++itr )
  {
    int dx = *itr;

    // there should be an even number of coordinates!
    ++itr;
    if( itr == foot->relativeCoordinates.end() )
    {
      printError( "Odd number of coordinates for footprint %s %s.\n",
                  type->data(),
                  name->data() );
      exit( -1 );
    }

    int dy = *itr;

    point dc;
    dc.pcSet( c->pcx + dx, c->pcy + dy );

    if( !isPlayableCell( &dc ) )
    {
      continue;
    }

    if( *type == "doodad" || *type == "unit" ) {
      // remove all pathing
      setPathing( &dc, PATH_GROUND_NOROCKS,               false );
      setPathing( &dc, PATH_GROUND_WITHROCKS,             false );
      setPathing( &dc, PATH_CWALK_NOROCKS,                false );
      setPathing( &dc, PATH_CWALK_WITHROCKS,              false );
      setPathing( &dc, PATH_GROUND_WITHROCKS_NORESOURCES, false );
      setPathing( &dc, PATH_BUILDABLE,                    false );

    } else if( *type == "destruct" ) {
      // remove from WITHROCKS pathing types
      setPathing( &dc, PATH_GROUND_WITHROCKS,             false );
      setPathing( &dc, PATH_CWALK_WITHROCKS,              false );
      setPathing( &dc, PATH_GROUND_WITHROCKS_NORESOURCES, false );

    } else if( *type == "resource" ) {
      // remove all pathing except WITHOUTRESOURCES
      setPathing( &dc, PATH_GROUND_NOROCKS,               false );
      setPathing( &dc, PATH_GROUND_WITHROCKS,             false );
      setPathing( &dc, PATH_CWALK_NOROCKS,                false );
      setPathing( &dc, PATH_CWALK_WITHROCKS,              false );
      setPathing( &dc, PATH_BUILDABLE,                    false );

    } else if( *type == "nobuild" ) {
      setPathing( &dc, PATH_BUILDABLE,                    false );

    } else {
      printError( "Placed object %s has unknown type %s.",
                  name->data(),
                  type->data() );
      exit( -1 );
    }
  }
}
